library static_sharedpreferences;

import 'dart:async';

import 'package:shared_preferences/shared_preferences.dart';

///For static use of the SharedPreferences Instance
class SP {
  static SharedPreferences get sharedPrefs =>
      (_sp != null) ? _sp : onSharedPreferencesIsNull("get sharedPrefs");
  static SharedPreferences _sp;

  static bool get isInitialized => _isInitializedSP;
  static bool _isInitializedSP = false;

  static StreamController<SPData<int>> _intStreamController;
  static StreamController<SPData<double>> _doubleStreamController;
  static StreamController<SPData<String>> _stringStreamController;
  static StreamController<SPData<List<String>>> _stringListStreamController;
  static StreamController<SPData<dynamic>> _dynamicStreamController;
  static StreamController<SPData<bool>> _boolStreamController;

  static Stream<SPData<int>> get intStream => _intStreamController.stream;
  static Stream<SPData<double>> get doubleStream =>
      _doubleStreamController.stream;
  static Stream<SPData<String>> get stringStream =>
      _stringStreamController.stream;
  static Stream<SPData<List<String>>> get stringListStream =>
      _stringListStreamController.stream;
  static Stream<SPData<dynamic>> get dynamicStream =>
      _dynamicStreamController.stream;
  static Stream<SPData<bool>> get boolStream => _boolStreamController.stream;

  static bool get isInitializedStreams => _isInitializedStreams;
  static bool _isInitializedStreams = false;

  /// ## Description
  ///
  /// * You have to put this method in your main(), like in the example down below.
  ///
  ///
  /// * Loads and parses the [SharedPreferences] for this app from disk.
  ///   Because this is reading from disk, it shouldn't be awaited in
  ///   performance-sensitive blocks.
  ///
  ///
  /// * Returns whether SP is initialized.
  ///
  ///
  /// * If you depend on a package which depends on `static_sharedpreferences`
  ///   then use [initializeOther] instead of [initialize]
  ///
  ///
  /// ## Example
  /// {@tool sample}
  /// ```dart
  /// void main() {
  ///   WidgetsFlutterBinding.ensureInitialized();
  ///
  ///   SP.initialize().whenComplete((){
  ///     print('Done with initialization!');
  ///   });
  /// }
  ///
  /// ```
  /// {@end-tool}
  static Future<bool> initialize({
    bool syncStreamInt = true,
    bool syncStreamDouble = true,
    bool syncStreamString = true,
    bool syncStreamStringList = true,
    bool syncStreamDynamic = true,
    bool syncStreamBool = true,
  }) async {
    if (!_isInitializedStreams) {
      _intStreamController = StreamController.broadcast(sync: syncStreamInt);
      _doubleStreamController =
          StreamController.broadcast(sync: syncStreamDouble);
      _stringStreamController =
          StreamController.broadcast(sync: syncStreamString);
      _stringListStreamController =
          StreamController.broadcast(sync: syncStreamStringList);
      _dynamicStreamController =
          StreamController.broadcast(sync: syncStreamDynamic);
      _boolStreamController = StreamController.broadcast(sync: syncStreamBool);

      _isInitializedStreams = true;
    }
    if (!_isInitializedSP) {
      _sp = await SharedPreferences.getInstance();
      (_sp != null)
          ? _isInitializedSP = true
          : onSharedPreferencesIsNull(
              "initialize, isInitialized: $_isInitializedSP");
    }

    return _isInitializedSP;
  }

  /// ## Description
  ///
  /// * In case you are not sure if SP has already initialized,
  ///   then you can use this method to save or restore some values.
  ///
  ///
  /// * In case SP still can not initialize because of some
  ///   unknown error `onSharedPreferencesIsNull` gets executed.
  ///   Otherwise this method is your way to go to save or restore values.
  ///
  ///
  /// * Returns [isInitialized].
  static Future<bool> execute(Function task) async {
    if (!_isInitializedSP) {
      await SP.initialize();
      task();
    } else {
      task();
    }
    return _isInitializedSP;
  }

  /// ## Description
  ///
  /// * Returns true, if first time launched, but before returning
  ///   a value,`onFirstTimeLaunched()` or `onNotFirstTimeLaunched()`
  ///   get executed if they are defined.
  static bool onFirstTimeLaunched({
    void Function() onFirstTimeLaunched,
    void Function() onNotFirstTimeLaunched,
  }) {
    if (_sp != null) {
      bool firstTimeLaunched = getBool(_firstTimeLaunchedKey) ?? true;
      if (firstTimeLaunched || firstTimeLaunched == null) {
        //First time launched
        setBool(_firstTimeLaunchedKey, false);
        if (onFirstTimeLaunched != null) onFirstTimeLaunched();
        return true;
      } else {
        //not first time launched
        if (onNotFirstTimeLaunched != null) onNotFirstTimeLaunched();
        return false;
      }
    } else {
      return onSharedPreferencesIsNull("onFirstTimeLaunched");
    }
  }

  /// Removes all values in persistant storage
  static void removeAll() {
    if (_sp != null) {
      Set<String> keys = _sp.getKeys();
      for (var item in keys) {
        _sp.remove(item);
      }
    } else {
      onSharedPreferencesIsNull("removeAll");
    }
  }

  /// Returns all keys in the persistent storage.
  static Set<String> getKeys() =>
      (_sp != null) ? _sp.getKeys() : onSharedPreferencesIsNull("getKeys");

  /// Reads a value from persistent storage, throwing an exception if it's not an int.
  static int getInt(String key) {
    if (_sp != null) {
      final val = _sp.getInt(key);
      _intStreamController.add(SPData._(
        data: val,
        dataKey: key,
        methodType: SPMethodType.getter,
      ));
      return val;
    } else {
      _intStreamController.add(SPData._(
        dataKey: key,
        methodType: SPMethodType.getter,
        hasError: true,
        errorData: "SharedPreferences is null",
      ));
      return onSharedPreferencesIsNull("getInt($key)");
    }
  }

  /// Saves an integer [value] to persistent storage in the background.
  /// If [value] is null, this is equivalent to calling [remove()] on the [key].
  static Future<bool> setInt(String key, int value) {
    if (_sp != null) {
      _intStreamController.add(SPData._(
        data: value,
        dataKey: key,
        methodType: SPMethodType.setter,
      ));
      return _sp.setInt(key, value);
    } else {
      _intStreamController.add(SPData._(
        data: value,
        dataKey: key,
        methodType: SPMethodType.setter,
        hasError: true,
        errorData: "SharedPreferences is null",
      ));
      return onSharedPreferencesIsNull("setInt($key, $value");
    }
  }

  /// Reads a value from persistent storage, throwing an exception if it's not a double.
  static double getDouble(String key) {
    if (_sp != null) {
      final val = _sp.getDouble(key);
      _doubleStreamController.add(SPData._(
        data: val,
        dataKey: key,
        methodType: SPMethodType.getter,
      ));
      return val;
    } else {
      _doubleStreamController.add(SPData._(
        dataKey: key,
        methodType: SPMethodType.getter,
        hasError: true,
        errorData: "SharedPreferences is null",
      ));
      return onSharedPreferencesIsNull("getDouble($key)");
    }
  }

  /// Saves a double [value] to persistent storage in the background.
  ///
  /// Android doesn't support storing doubles, so it will be stored as a float.
  ///
  /// If [value] is null, this is equivalent to calling [remove()] on the [key].
  static Future<bool> setDouble(String key, double value) {
    if (_sp != null) {
      _doubleStreamController.add(SPData._(
        data: value,
        dataKey: key,
        methodType: SPMethodType.setter,
      ));
      return _sp.setDouble(key, value);
    } else {
      _doubleStreamController.add(SPData._(
        data: value,
        dataKey: key,
        methodType: SPMethodType.setter,
        hasError: true,
        errorData: "SharedPreferences is null",
      ));
      return onSharedPreferencesIsNull("setDouble($key, $value");
    }
  }

  /// Reads a value from persistent storage, throwing an exception if it's not a String.
  static String getString(String key) {
    if (_sp != null) {
      final val = _sp.getString(key);
      _stringStreamController.add(SPData._(
        data: val,
        dataKey: key,
        methodType: SPMethodType.getter,
      ));
      return val;
    } else {
      _stringStreamController.add(SPData._(
        dataKey: key,
        methodType: SPMethodType.getter,
        hasError: true,
        errorData: "SharedPreferences is null",
      ));
      return onSharedPreferencesIsNull("getString($key)");
    }
  }

  /// Saves a string [value] to persistent storage in the background.
  ///
  /// If [value] is null, this is equivalent to calling [remove()] on the [key].
  static Future<bool> setString(String key, String value) {
    if (_sp != null) {
      _stringStreamController.add(SPData._(
        data: value,
        dataKey: key,
        methodType: SPMethodType.setter,
      ));
      return _sp.setString(key, value);
    } else {
      _stringStreamController.add(SPData._(
        data: value,
        dataKey: key,
        methodType: SPMethodType.setter,
        hasError: true,
        errorData: "SharedPreferences is null",
      ));
      return onSharedPreferencesIsNull("setString($key, $value)");
    }
  }

  /// Reads a set of string values from persistent storage, throwing an exception if it's not a string set.
  static List<String> getStringList(String key) {
    if (_sp != null) {
      final val = _sp.getStringList(key);
      _stringListStreamController.add(SPData._(
        data: val,
        dataKey: key,
        methodType: SPMethodType.getter,
      ));
      return val;
    } else {
      _stringListStreamController.add(SPData._(
        dataKey: key,
        methodType: SPMethodType.getter,
        hasError: true,
        errorData: "SharedPreferences is null",
      ));
      return onSharedPreferencesIsNull("getStringList($key)");
    }
  }

  /// Saves a list of strings [value] to persistent storage in the background.
  ///
  /// If [value] is null, this is equivalent to calling [remove()] on the [key].
  static Future<bool> setStringList(String key, List<String> value) {
    if (_sp != null) {
      _stringListStreamController.add(SPData._(
        data: value,
        dataKey: key,
        methodType: SPMethodType.setter,
      ));
      return _sp.setStringList(key, value);
    } else {
      _stringListStreamController.add(SPData._(
        data: value,
        dataKey: key,
        methodType: SPMethodType.setter,
        hasError: true,
        errorData: "SharedPreferences is null",
      ));
      return onSharedPreferencesIsNull("setStringList($key, $value)");
    }
  }

  /// Reads a value from persistent storage, throwing an exception if it's not a bool.
  static bool getBool(String key) {
    if (_sp != null) {
      final val = _sp.getBool(key);
      _boolStreamController.add(SPData._(
        data: val,
        dataKey: key,
        methodType: SPMethodType.getter,
      ));
      return val;
    } else {
      _boolStreamController.add(SPData._(
        dataKey: key,
        methodType: SPMethodType.getter,
        hasError: true,
        errorData: "SharedPreferences is null",
      ));
      return onSharedPreferencesIsNull("getBool($key)");
    }
  }

  /// Saves a boolean [value] to persistent storage in the background.
  ///
  /// If [value] is null, this is equivalent to calling [remove()] on the [key].
  static Future<bool> setBool(String key, bool value) {
    if (_sp != null) {
      _boolStreamController.add(SPData._(
        data: value,
        dataKey: key,
        methodType: SPMethodType.setter,
      ));
      return _sp.setBool(key, value);
    } else {
      _boolStreamController.add(SPData._(
        data: value,
        dataKey: key,
        methodType: SPMethodType.setter,
        hasError: true,
        errorData: "SharedPreferences is null",
      ));
      return onSharedPreferencesIsNull("setBool($key, $value)");
    }
  }

  /// Reads a value of any type from persistent storage.
  static dynamic getDynamic(String key) {
    if (_sp != null) {
      final val = _sp.get(key);
      _dynamicStreamController.add(SPData._(
        data: val,
        dataKey: key,
        methodType: SPMethodType.getter,
      ));
      return val;
    } else {
      _dynamicStreamController.add(SPData._(
        dataKey: key,
        methodType: SPMethodType.getter,
        hasError: true,
        errorData: "SharedPreferences is null",
      ));
      return onSharedPreferencesIsNull("getDynamic($key)");
    }
  }

  /// ## Description
  ///
  /// * If you override this make sure to `return returnValue`.
  ///
  ///
  /// * This method gets triggered if you try to use a method,
  ///   which required the SharedPreferences instance, but
  ///   the instance == null.
  ///
  ///
  /// * If not changed by the user, then this will just use
  ///   the print() function to notify the developer
  ///   of the state of the SharedPreferences instance.
  ///
  /// ## Example
  ///
  /// {@tool sample}
  /// ```dart
  /// //some code...
  ///
  /// int value = SP.getInt(myKey);
  /// //You try to restore an integer,
  /// //but in case SP.sharedPrefs == null
  /// //then onSharedPreferencesIsNull() gets executed
  ///
  /// //some code...
  /// ```
  /// {@end-tool}
  static OnSharedPreferencesIsNullFunction onSharedPreferencesIsNull =
      (String event, {dynamic returnValue}) {
    print("\n\n\n SharedPreferences is $_sp \n Triggered on: $event \n\n");
    return returnValue;
  };
}

/// ## Description
///
/// * Can be ideally used for switch- or if-statements
///
/// ## Example
///
/// {@tool sample}
/// ```dart
/// switch (mySpType) {
///   case SP_Type.BOOL:
///     SP.setBool(myKey, true);
///     break;
///   case SP_Type.INT:
///     int cache = SP.getInt(myKey) ?? 0;
///     SP.setInt(myKey, cache + 1);
///     break;
///   default:
/// }
///
/// ```
///
/// {@end-tool}
enum SP_Type {
  INT,
  DOUBLE,
  STRING,
  BOOL,
  DYNAMIC,
  LIST_STRING,
}

typedef OnSharedPreferencesIsNullFunction = dynamic Function(String event,
    {dynamic returnValue});

const String _firstTimeLaunchedKey =
    "static_sharedpreferences;SP.onFirstTimeLaunched:oasj914A5I2SC318";

class SPData<T> {
  final T data;
  final String dataKey;
  final SPMethodType methodType;
  final bool hasError;
  final dynamic errorData;

  const SPData._({
    this.data,
    this.dataKey,
    this.methodType,
    this.hasError = false,
    this.errorData,
  });
}

enum SPMethodType { getter, setter }
